import string, re, Globals
from Globals             import DTMLFile
from DateTime            import DateTime
from OFS.SimpleItem      import SimpleItem
from OFS.PropertyManager import PropertyManager

logicKeywords = {'and' : lambda x, y: x and y,
                 'or'  : lambda x, y: x or y,
                 'not' : lambda x, y: not y}
termParser = re.compile('\s*(".*"|%s)\s*' %(string.join(logicKeywords.keys(), '|')))


manage_addSiteSearchForm = DTMLFile('dtml/manage_addSiteSearchForm', globals())

def manage_addSiteSearch(self, id, REQUEST=None):
    ''' '''
    ss = SiteSearch(id, 'Site Search Object')
    self._setObject(id, ss)

    if REQUEST:
        return self.manage_main(self, REQUEST)



class SiteSearch(PropertyManager, SimpleItem):
    ''' '''

    meta_type = 'Site Search'

    __ac_permissions__=( ( 'View', ('findObjects',) ),
                         ( 'View Management Screens', ('manage_searchSiteForm', 'manage_searchSite') ),
                         )

    manage_options = ( ( {'label':'Search', 'action':'manage_searchSiteForm'}, ) 
                       + PropertyManager.manage_options
                       + SimpleItem.manage_options )

    _properties = (
        {'id': 'objectMetaTypes',        'type': 'lines',   'mode': 'w', 'default': ['all']},
        {'id': 'objectModTime',          'type': 'string',  'mode': 'w', 'default': ''},
        {'id': 'objectModSpecs',         'type': 'string',  'mode': 'w', 'default': ''},
        {'id': 'searchSubObjects',       'type': 'boolean', 'mode': 'w', 'default': 1},
        {'id': 'searchControlPanel',     'type': 'boolean', 'mode': 'w', 'default': 0},
        {'id': 'searchObjectProperties', 'type': 'boolean', 'mode': 'w', 'default': 1},
                   )

    manage_searchSiteForm = DTMLFile('dtml/manage_searchSiteForm', globals())
    manage_searchSite     = DTMLFile('dtml/manage_searchSite', globals())


    def __init__(self, id, title = '', **kw):
        ''' '''
        self.id = str(id)
        self.title = title

        for property in self._properties:
            if property['id'] in kw.keys():
                setattr(self, property['id'], kw[property['id']])
            else:
                setattr(self, property['id'], property['default'])

    
    def findObjects(self,
                    baseObject       = None,   # Define the object where the search should start
                    objectIds        = None,   # search only for these object ids 
                    objectMetaTypes  = None,   # object must be one of meta-types
                    objectSearchTerm = None,   # search for these conditions
                    objectModTime    = None,   # object modification time
                    objectModSpecs   = None,   # whether to search before/after mod. time
                    searchSubObjects = 0,      # should we also look into sub objects?
                    result           = None,   # Just in case there are already some results
                    prefix           = ''):
        ''' '''

        # Since I noticed that it can become very annoying, that the search also looks into
        # the Control Panel and everything below, I simply made it an option now
        if baseObject.getId() == 'Control_Panel' and self.searchControlPanel == 'off':
            return result

        if result is None:
            result = []

            # Yeah, yeah, I know, it's clumsy, but somehow it would not work with locals()
            if objectMetaTypes is None:
                objectMetaTypes = getattr(self, 'objectMetaTypes')

            if objectModTime is None:
                objectModTime = getattr(self, 'objectModTime')

            if objectModSpecs is None:
                objectModSpecs = getattr(self, 'objectModSpecs')

            if searchSubObjects == 0:
                if getattr(self, 'searchSubObjects') == 'on':
                    searchSubObjects = 1


            if ( (objectMetaTypes and 'all' in objectMetaTypes) or
                 filter(None, objectMetaTypes) == [] ):
                objectMetaTypes = None
                
            if objectModTime and type(objectModTime) == type(''):
                objectModTime = DateTime(objectModTime).timeTime()

            if objectSearchTerm and type(objectSearchTerm) == type(''):
                objectSearchTerm = self._parseSearchTerm(objectSearchTerm)


        # set the base object of the search; if none was specified use the parent of the
        # search object itself.
        if baseObject:
            searchBase = baseObject
            if hasattr(baseObject, 'aq_base'):
                searchBase = baseObject.aq_base
        else:
            searchBase = self.aq_parent


        # Now try to see whether we can get sub-objects of this one. That means that
        # this object should be folderish. If not return the current result list of
        # objects
        if not hasattr(searchBase, 'objectItems'):
            return result

        # You can never be sure it works.
        try:
            items = baseObject.objectItems()
        except:
            return result


        # Now make sure that the result is a list-like object and supports append
        try:
            add_result = result.append
        except:
            raise AttributeError, `result`

        # Okay, we are finally ready to search. Yeah! Since items contains an (id, object) pair,
        # we can unpack them. I dunno how much good that does, but it might be a speed increase
        # compared to looking up the id all the time.
        for id, object in items:
                
            # Okay, since we are keeping track of the path, we need to work on the prefix by
            # updating it
            if prefix:
                newPrefix = "%s/%s" %(prefix, id)
            else:
                newPrefix = id

            # Has anyone a clue, why this is done? And what does dflag stand for?
            dflag = 0
            if hasattr(object, '_p_changed') and (object._p_changed == None):
                dflag = 1

            
            if hasattr(object, 'aq_base'):
                objectBase = object.aq_base
            else:
                objectBase = object


            if ( ( not objectIds or absattr(objectBase.getId()) in objectIds) and
                 
                 ( not objectMetaTypes or ( hasattr(objectBase, 'meta_type') and
                                            objectBase.meta_type in objectMetaTypes ) ) and

                 ( not objectSearchTerm or ( ( hasattr(object, 'PrincipiaSearchSource') or
                                               hasattr(object, 'propertyValues') ) and
                                             self._searchTermInObject(object, objectSearchTerm)) ) and

                 ( not objectModTime or self._modTimeMatch(object, objectModTime, objectModSpecs) )

                ):
                add_result((newPrefix, object))
                dflag = 0
                    
            if searchSubObjects and hasattr(objectBase, 'objectItems'):
                self.findObjects(object,
                                 objectIds,
                                 objectMetaTypes,
                                 objectSearchTerm,
                                 objectModTime,
                                 objectModSpecs,
                                 searchSubObjects,
                                 result, newPrefix)
                
            if dflag:
                object._p_deactivate()

        return result


    
    def _modTimeMatch(self, object, modTime, quaifier, function=hasattr):
        if not function(object, '_p_mtime'):
            return 0    
        return qualifier == '<' and (object._p_mtime < modTime) or (object._p_mtime > modTime)
    
    
    def _parseSearchTerm(self, term):
        res = termParser.split(term)
        res = filter(None, res)
        res = map(lambda x, r=string.replace: r(x, '"', ''), res)
        return res
    
    
    def _searchTermInObject(self, object, term):
    
        # Geez, we need some docu for this method. But it is fairly nice, since you just need
        # to define any other logic keyword on the top and the method will still work.
    
        previous  = 1     # keeps track on how the previous phrase evaluated
        current   = 1     # keeps track on how the currect phrase evaluated (was found or not)
        operation = 'and' # The operation that is used to compare previous and current
        
        # for every phrase and operation in the term (which is already a list a this point)
        for phrase in term:
            # check whether the phrase is a keyword; if so change the operation
            if phrase not in logicKeywords.keys():
    
                # setup the variables
                prinResult = 0
                propResult = 0
                
                # see whether the phrase exist in the object principia search source
                if hasattr(object, 'PrincipiaSearchSource'):
                    prinResult = string.find(object.PrincipiaSearchSource(), phrase) + 1
    
                # if we want, also check through the attributes
                if self.searchObjectProperties == 'on':
                    propResult = self._searchThroughProperties(object, phrase)
    
                # now get the total result
                current = prinResult or propResult
    
                # now execute the logic
                previous = logicKeywords[operation](previous, current)
    
            else:
                operation = phrase
    
            if not previous:
                return 0
            
        return 1
            
    
    def _searchThroughProperties(self, object, phrase):
    
        if hasattr(object, 'propertyValues'):
            for value in object.propertyValues():
                if string.find(str(value), phrase) >= 0:
                    return 1
        return 0


Globals.default__class_init__(SiteSearch)



# Helper functions

def absattr(attr):
    if callable(attr):
        return attr()
    return attr
